@@ -2,8 +2,14 @@ import distutils.util
 import importlib
 import inspect
 import shlex
+from typing import Iterable
 
 from jb_declarative_formatters.parsers.type_name_parser import parse_type_name_template
+from jb_declarative_formatters.type_viz_synthetic_method import SyntheticMethodDefinition
+from jb_lldb_polyfills import \
+    LLDBRemoveAllTopLevelLazyDeclarations, \
+    LLDBAddTopLevelLazyDeclarationByRegex, \
+    LLDBAddTopLevelLazyDeclaration
 from renderers.jb_lldb_builtin_formatters import *
 from renderers.jb_lldb_declarative_formatters_loaders import *
 from renderers.jb_lldb_declarative_formatters_manager import *
@@ -28,6 +34,7 @@ def __lldb_init_module(debugger: lldb.SBDebugger, internal_dict):
 
         make_absolute_name(__name__, '_cmd_reload_all'): 'jb_renderers_reload_all',
         make_absolute_name(__name__, '_cmd_remove_all'): 'jb_renderers_remove_all',
+        make_absolute_name(__name__, '_cmd_list_all'):   'jb_renderers_list_all',
 
         make_absolute_name(__name__, '_cmd_override_charset'): 'jb_renderers_override_charset',
         make_absolute_name(__name__, '_cmd_set_markup'): 'jb_renderers_set_markup',
@@ -121,7 +128,8 @@ def _cmd_load(debugger, command, exe_ctx, result, internal_dict):
     file_paths = cmd[1:]
     for filepath in file_paths:
         try:
-            lldb_formatters_manager.register(filepath, loader)
+            entry = lldb_formatters_manager.register(filepath, loader)
+            add_all_top_level_declarations(debugger, [entry])
         except TypeVizLoaderException as e:
             result.SetError('{}'.format(str(e)))
             return
@@ -155,6 +163,10 @@ def _cmd_reload_all(debugger, command, exe_ctx, result, internal_dict):
     reload_all(debugger)
 
 
+def _cmd_list_all(debugger, command, exe_ctx, result, internal_dict):
+    result.AppendMessage("\n".join(get_all_registered_files()))
+
+
 def _cmd_override_charset(debugger, command, exe_ctx, result, internal_dict):
     help_message = 'Usage: jb_renderers_override_charset <charset>'
     cmd = shlex.split(command)
@@ -209,14 +221,48 @@ def reload_all(debugger):
     reload_file_list(debugger, files)
 
 
+def get_all_registered_files():
+    return lldb_formatters_manager.get_all_registered_files()
+
+
 def remove_file_list(debugger, files):
     for filepath in files:
         lldb_formatters_manager.unregister(filepath)
+    LLDBRemoveAllTopLevelLazyDeclarations(debugger)
+    add_all_top_level_declarations(debugger, lldb_formatters_manager.formatter_entries.values())
 
 
 def reload_file_list(debugger, files):
     for filepath in files:
         lldb_formatters_manager.reload(filepath)
+    LLDBRemoveAllTopLevelLazyDeclarations(debugger)
+    add_all_top_level_declarations(debugger, lldb_formatters_manager.formatter_entries.values())
+
+
+def add_all_top_level_declarations(debugger: lldb.SBDebugger, entries: Iterable[FormattersManager.FormatterEntry]):
+    for entry in entries:
+        top_level_methods: List[SyntheticMethodDefinition] = entry.storage.get_top_level_methods()
+        for top_level_method_definition in top_level_methods:
+            if top_level_method_definition.name_uses_regex:
+                error: lldb.SBError = LLDBAddTopLevelLazyDeclarationByRegex(
+                    debugger,
+                    top_level_method_definition.full_name,
+                    top_level_method_definition.body_substitution,
+                    eLanguageTypeC_plus_plus_14
+                )
+                if not error.Success():
+                    log(f"Can't add lazy declarations by regex for '{top_level_method_definition.full_name}': "
+                        f"{error.description}")
+            else:
+                error: lldb.SBError = LLDBAddTopLevelLazyDeclaration(
+                    debugger,
+                    top_level_method_definition.full_name,
+                    top_level_method_definition.body_substitution,
+                    eLanguageTypeC_plus_plus_14
+                )
+                if not error.Success():
+                    log(f"Can't add lazy declarations for '{top_level_method_definition.full_name}': "
+                        f"{error.description}")
 
 
 def declarative_summary(val: lldb.SBValue, internal_dict):
